package sun.reflect;

import java.util.Objects;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.OptionalDataException;
import java.lang.reflect.InvocationTargetException;
import java.io.ObjectStreamClass;
import java.lang.invoke.MethodHandles;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.lang.invoke.MethodHandle;
import java.io.Externalizable;
import java.io.Serializable;
import java.lang.reflect.Executable;
import java.lang.reflect.Modifier;
import java.lang.reflect.Constructor;
import sun.reflect.misc.ReflectUtil;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.security.Permission;

public class ReflectionFactory
{
    private static boolean initted;
    private static final Permission reflectionFactoryAccessPerm;
    private static final ReflectionFactory soleInstance;
    private static volatile LangReflectAccess langReflectAccess;
    private static volatile Method hasStaticInitializerMethod;
    private static boolean noInflation;
    private static int inflationThreshold;
    
    private ReflectionFactory() {
    }
    
    public static ReflectionFactory getReflectionFactory() {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(ReflectionFactory.reflectionFactoryAccessPerm);
        }
        return ReflectionFactory.soleInstance;
    }
    
    public void setLangReflectAccess(LangReflectAccess langReflectAccess) {
        ReflectionFactory.langReflectAccess = langReflectAccess;
    }
    
    public FieldAccessor newFieldAccessor(Field field, boolean b) {
        checkInitted();
        return UnsafeFieldAccessorFactory.newFieldAccessor(field, b);
    }
    
    public MethodAccessor newMethodAccessor(Method method) {
        checkInitted();
        if (ReflectionFactory.noInflation && !ReflectUtil.isVMAnonymousClass(method.getDeclaringClass())) {
            return new MethodAccessorGenerator().generateMethod(method.getDeclaringClass(), method.getName(), method.getParameterTypes(), method.getReturnType(), method.getExceptionTypes(), method.getModifiers());
        }
        NativeMethodAccessorImpl nativeMethodAccessorImpl = new NativeMethodAccessorImpl(method);
        DelegatingMethodAccessorImpl parent = new DelegatingMethodAccessorImpl(nativeMethodAccessorImpl);
        nativeMethodAccessorImpl.setParent(parent);
        return parent;
    }
    
    public ConstructorAccessor newConstructorAccessor(Constructor<?> constructor) {
        checkInitted();
        Class<Object> declaringClass = constructor.getDeclaringClass();
        if (Modifier.isAbstract(declaringClass.getModifiers())) {
            return new InstantiationExceptionConstructorAccessorImpl(null);
        }
        if (declaringClass == Class.class) {
            return new InstantiationExceptionConstructorAccessorImpl("Can not instantiate java.lang.Class");
        }
        if (Reflection.isSubclassOf(declaringClass, ConstructorAccessorImpl.class)) {
            return new BootstrapConstructorAccessorImpl(constructor);
        }
        if (ReflectionFactory.noInflation && !ReflectUtil.isVMAnonymousClass(constructor.getDeclaringClass())) {
            return new MethodAccessorGenerator().generateConstructor(constructor.getDeclaringClass(), constructor.getParameterTypes(), constructor.getExceptionTypes(), constructor.getModifiers());
        }
        NativeConstructorAccessorImpl nativeConstructorAccessorImpl = new NativeConstructorAccessorImpl(constructor);
        DelegatingConstructorAccessorImpl parent = new DelegatingConstructorAccessorImpl(nativeConstructorAccessorImpl);
        nativeConstructorAccessorImpl.setParent(parent);
        return parent;
    }
    
    public Field newField(Class<?> clazz, String s, Class<?> clazz2, int n, int n2, String s2, byte[] array) {
        return langReflectAccess().newField(clazz, s, clazz2, n, n2, s2, array);
    }
    
    public Method newMethod(Class<?> clazz, String s, Class<?>[] array, Class<?> clazz2, Class<?>[] array2, int n, int n2, String s2, byte[] array3, byte[] array4, byte[] array5) {
        return langReflectAccess().newMethod(clazz, s, array, clazz2, array2, n, n2, s2, array3, array4, array5);
    }
    
    public Constructor<?> newConstructor(Class<?> clazz, Class<?>[] array, Class<?>[] array2, int n, int n2, String s, byte[] array3, byte[] array4) {
        return langReflectAccess().newConstructor(clazz, array, array2, n, n2, s, array3, array4);
    }
    
    public MethodAccessor getMethodAccessor(Method method) {
        return langReflectAccess().getMethodAccessor(method);
    }
    
    public void setMethodAccessor(Method method, MethodAccessor methodAccessor) {
        langReflectAccess().setMethodAccessor(method, methodAccessor);
    }
    
    public ConstructorAccessor getConstructorAccessor(Constructor<?> constructor) {
        return langReflectAccess().getConstructorAccessor(constructor);
    }
    
    public void setConstructorAccessor(Constructor<?> constructor, ConstructorAccessor constructorAccessor) {
        langReflectAccess().setConstructorAccessor(constructor, constructorAccessor);
    }
    
    public Method copyMethod(Method method) {
        return langReflectAccess().copyMethod(method);
    }
    
    public Field copyField(Field field) {
        return langReflectAccess().copyField(field);
    }
    
    public <T> Constructor<T> copyConstructor(Constructor<T> constructor) {
        return langReflectAccess().copyConstructor(constructor);
    }
    
    public byte[] getExecutableTypeAnnotationBytes(Executable executable) {
        return langReflectAccess().getExecutableTypeAnnotationBytes(executable);
    }
    
    public Constructor<?> newConstructorForSerialization(Class<?> clazz, Constructor<?> constructor) {
        if (constructor.getDeclaringClass() == clazz) {
            return constructor;
        }
        return this.generateConstructor(clazz, constructor);
    }
    
    public final Constructor<?> newConstructorForSerialization(Class<?> superclass) {
        while (Serializable.class.isAssignableFrom(superclass)) {
            if ((superclass = superclass.getSuperclass()) != null) {
                continue;
            }
            return null;
        }
        Constructor<Object> declaredConstructor;
        try {
            declaredConstructor = superclass.getDeclaredConstructor(new Class[0]);
            int modifiers = declaredConstructor.getModifiers();
            if ((modifiers & 0x2) != 0x0 || ((modifiers & 0x5) == 0x0 && !packageEquals(superclass, superclass))) {
                return null;
            }
        }
        catch (NoSuchMethodException ex) {
            return null;
        }
        return this.generateConstructor(superclass, declaredConstructor);
    }
    
    private final Constructor<?> generateConstructor(Class<?> clazz, Constructor<?> constructor) {
        SerializationConstructorAccessorImpl generateSerializationConstructor = new MethodAccessorGenerator().generateSerializationConstructor(clazz, constructor.getParameterTypes(), constructor.getExceptionTypes(), constructor.getModifiers(), constructor.getDeclaringClass());
        Constructor<Object> constructor2 = this.newConstructor(constructor.getDeclaringClass(), constructor.getParameterTypes(), constructor.getExceptionTypes(), constructor.getModifiers(), langReflectAccess().getConstructorSlot(constructor), langReflectAccess().getConstructorSignature(constructor), langReflectAccess().getConstructorAnnotations(constructor), langReflectAccess().getConstructorParameterAnnotations(constructor));
        this.setConstructorAccessor(constructor2, generateSerializationConstructor);
        constructor2.setAccessible(true);
        return constructor2;
    }
    
    public final Constructor<?> newConstructorForExternalization(Class<?> clazz) {
        if (!Externalizable.class.isAssignableFrom(clazz)) {
            return null;
        }
        try {
            Constructor<Object> constructor = clazz.getConstructor(new Class[0]);
            constructor.setAccessible(true);
            return constructor;
        }
        catch (NoSuchMethodException ex) {
            return null;
        }
    }
    
    public final MethodHandle readObjectForSerialization(Class<?> clazz) {
        return this.findReadWriteObjectForSerialization(clazz, "readObject", ObjectInputStream.class);
    }
    
    public final MethodHandle readObjectNoDataForSerialization(Class<?> clazz) {
        return this.findReadWriteObjectForSerialization(clazz, "readObjectNoData", ObjectInputStream.class);
    }
    
    public final MethodHandle writeObjectForSerialization(Class<?> clazz) {
        return this.findReadWriteObjectForSerialization(clazz, "writeObject", ObjectOutputStream.class);
    }
    
    private final MethodHandle findReadWriteObjectForSerialization(Class<?> clazz, String name, Class<?> clazz2) {
        if (!Serializable.class.isAssignableFrom(clazz)) {
            return null;
        }
        try {
            Method declaredMethod = clazz.getDeclaredMethod(name, new Class[] { clazz2 });
            int modifiers = declaredMethod.getModifiers();
            if (declaredMethod.getReturnType() != Void.TYPE || Modifier.isStatic(modifiers) || !Modifier.isPrivate(modifiers)) {
                return null;
            }
            declaredMethod.setAccessible(true);
            return MethodHandles.lookup().unreflect(declaredMethod);
        }
        catch (NoSuchMethodException ex) {
            return null;
        }
        catch (IllegalAccessException cause) {
            throw new InternalError("Error", cause);
        }
    }
    
    public final MethodHandle readResolveForSerialization(Class<?> clazz) {
        return this.getReplaceResolveForSerialization(clazz, "readResolve");
    }
    
    public final MethodHandle writeReplaceForSerialization(Class<?> clazz) {
        return this.getReplaceResolveForSerialization(clazz, "writeReplace");
    }
    
    private MethodHandle getReplaceResolveForSerialization(Class<?> superclass, String name) {
        if (!Serializable.class.isAssignableFrom(superclass)) {
            return null;
        }
        while (superclass != null) {
            try {
                Method declaredMethod = superclass.getDeclaredMethod(name, new Class[0]);
                if (declaredMethod.getReturnType() != Object.class) {
                    return null;
                }
                int modifiers = declaredMethod.getModifiers();
                if ((Modifier.isStatic(modifiers) | Modifier.isAbstract(modifiers))) {
                    return null;
                }
                if (!(Modifier.isPublic(modifiers) | Modifier.isProtected(modifiers))) {
                    if (Modifier.isPrivate(modifiers) && superclass != superclass) {
                        return null;
                    }
                    if (!packageEquals(superclass, superclass)) {
                        return null;
                    }
                }
                try {
                    declaredMethod.setAccessible(true);
                    return MethodHandles.lookup().unreflect(declaredMethod);
                }
                catch (IllegalAccessException cause) {
                    throw new InternalError("Error", cause);
                }
            }
            catch (NoSuchMethodException ex) {
                superclass = superclass.getSuperclass();
                continue;
            }
            break;
        }
        return null;
    }
    
    public final boolean hasStaticInitializerForSerialization(Class<?> clazz) {
        Method hasStaticInitializerMethod = ReflectionFactory.hasStaticInitializerMethod;
        if (hasStaticInitializerMethod == null) {
            try {
                hasStaticInitializerMethod = ObjectStreamClass.class.getDeclaredMethod("hasStaticInitializer", new Class[] { Class.class });
                hasStaticInitializerMethod.setAccessible(true);
                ReflectionFactory.hasStaticInitializerMethod = hasStaticInitializerMethod;
            }
            catch (NoSuchMethodException cause) {
                throw new InternalError("No such method hasStaticInitializer on " + ObjectStreamClass.class, cause);
            }
        }
        try {
            return (boolean)hasStaticInitializerMethod.invoke(null, new Object[] { clazz });
        }
        catch (InvocationTargetException | IllegalAccessException cause2) {
            throw new InternalError("Exception invoking hasStaticInitializer", cause2);
        }
    }
    
    public final OptionalDataException newOptionalDataExceptionForSerialization(boolean b) {
        try {
            Constructor<OptionalDataException> declaredConstructor = OptionalDataException.class.getDeclaredConstructor(new Class[] { Boolean.TYPE });
            declaredConstructor.setAccessible(true);
            return (OptionalDataException)declaredConstructor.newInstance(new Object[] { Boolean.valueOf(b) });
        }
        catch (NoSuchMethodException | InstantiationException | IllegalAccessException | InvocationTargetException cause) {
            throw new InternalError("unable to create OptionalDataException", cause);
        }
    }
    
    static int inflationThreshold() {
        return ReflectionFactory.inflationThreshold;
    }
    
    private static void checkInitted() {
        if (ReflectionFactory.initted) {
            return;
        }
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                if (System.out == null) {
                    return null;
                }
                String property = System.getProperty("sun.reflect.noInflation");
                if (property != null && property.equals("true")) {
                    ReflectionFactory.noInflation = true;
                }
                String property2 = System.getProperty("sun.reflect.inflationThreshold");
                if (property2 != null) {
                    try {
                        ReflectionFactory.inflationThreshold = Integer.parseInt(property2);
                    }
                    catch (NumberFormatException cause) {
                        throw new RuntimeException("Unable to parse property sun.reflect.inflationThreshold", cause);
                    }
                }
                ReflectionFactory.initted = true;
                return null;
            }
        });
        return;
    }
    
    private static LangReflectAccess langReflectAccess() {
        if (ReflectionFactory.langReflectAccess == null) {
            Modifier.isPublic(1);
        }
        return ReflectionFactory.langReflectAccess;
    }
    
    private static boolean packageEquals(Class<?> clazz, Class<?> clazz2) {
        return clazz.getClassLoader() == clazz2.getClassLoader() && Objects.equals(clazz.getPackage(), clazz2.getPackage());
    }
    
    static {
        ReflectionFactory.initted = false;
        reflectionFactoryAccessPerm = new RuntimePermission("reflectionFactoryAccess");
        soleInstance = new ReflectionFactory();
        ReflectionFactory.noInflation = false;
        ReflectionFactory.inflationThreshold = 15;
    }
    
    public static final class GetReflectionFactoryAction implements PrivilegedAction<ReflectionFactory>
    {
        @Override
        public ReflectionFactory run() {
            return ReflectionFactory.getReflectionFactory();
        }
    }
}
